/* Table for storing sequences represented by an LZW code */
#include <hammer/hammer.h>
#include <hammer/glue.h>
// malloc, free
#include <stdlib.h>
// strlen
#include <string.h>

#include "lzw.h"

//FILE *debug; // DEBUG

#define BITLIMIT_9 (ctx->earlychange ? 511 : 512)
#define BITLIMIT_10 (ctx->earlychange ? 1023 : 1024)
#define BITLIMIT_11 (ctx->earlychange ? 2047 : 2048)
/*
 *  Since bit lengths larger than 12 aren't allowed, EarlyChange doesn't matter here.
 */
#define BITLIMIT_12 4096


void LZW_clear_table(LZW_context_T *ctx)
{
	/*
	 *  Optimizations: since we leave the entries 0-257 fixed or empty, we don't need to free() them explicitly.
	 *  And since codes are added to the table sequentially, we don't need to look past ctx->next;
	 */
	for(int i = 258; i < ctx->next; ++i)
	{
		HBytes * sequence = ctx->lzw_code_table[i];
		if(sequence != NULL)
		{
			/*
			 * Assumption: only the HBytes in the LZW table refer to the particular uint8_t arrays we're freeing.
			 */
			free((uint8_t *)sequence->token);
			free(sequence);
		}
		ctx->lzw_code_table[i] = NULL;
	}
}

/*
 * Creates a HBytes from an array of bytes and its length, and inserts it into the lzw dictionary in ctx.
 * Also increments ctx->next. The HBytes will keep the token pointer, to be freed later in lzw_clear_table or init_lzw_context.
 */
void lzw_table_insert(LZW_context_T *ctx, uint8_t *token, size_t token_len)
{
	HBytes * next_entry = malloc(sizeof(HBytes));
	next_entry->token = token;
	next_entry->len = token_len;
	ctx->lzw_code_table[ctx->next] = next_entry;
	ctx->next++;
}

HParser *p_lzwdata;
LZW_context_T * context;

/*
 * First "code" in input. We output it literally, and set "old"
 */
HParsedToken*
act_LZW_firstcode(const HParseResult *p, void *u)
{
	/*HBytes * next_entry;
	size_t next_entry_size;
	uint8_t * next_entry_token;*/
	LZW_context_T * ctx = (LZW_context_T *) u;
	uint64_t code = H_CAST_UINT(p->ast);
	uint8_t *output = H_ALLOC(uint8_t);
	*output = (uint8_t) code;
	//fprintf(debug, "firstcode code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", p->ast->uint); // DEBUG
	//fflush(debug); // DEBUG

	ctx->old = code;
	return H_MAKE_BYTES(output, 1);
}

HParsedToken*
act_LZW_clear(const HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	LZW_clear_table(ctx);
	//fprintf(debug, "clear code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", p->ast->uint); // DEBUG
	//fflush(debug); // DEBUG
	ctx->next = 258; // Caution: moving this before the call to LZW_clear_table() will cause a memory leak
	return H_MAKE_BYTES(NULL, 0);
}

HParsedToken*
act_LZW_clear_and_firstcode(const HParseResult *p, void *u)
{
	return p->ast->seq->elements[1];
}

bool
validate_LZW_9bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "9 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	if (ctx->next < BITLIMIT_9) // DEBUG
		assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next < BITLIMIT_9);
}

bool
validate_LZW_10bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "10 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	if (ctx->next >= BITLIMIT_9 && ctx->next < BITLIMIT_10) // DEBUG
		assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_9 && ctx->next < BITLIMIT_10);
}

bool
validate_LZW_11bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "11 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	if (ctx->next >= BITLIMIT_10 && ctx->next < BITLIMIT_11) // DEBUG
		assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_10 && ctx->next < BITLIMIT_11);
}

bool
validate_LZW_12bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "12 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	if (ctx->next >= BITLIMIT_11 && ctx->next < BITLIMIT_12) // DEBUG
		assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_11 && ctx->next < BITLIMIT_12);
}

bool
validate_LZW_9bitlitspec(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "9 bit lit: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	uint64_t code = H_CAST_UINT(p->ast);
	return (ctx->next < BITLIMIT_9 && code < 258);
}

bool
validate_LZW_10bitlitspec(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "10 bit lit: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	uint64_t code = H_CAST_UINT(p->ast);
	return (ctx->next >= BITLIMIT_9 && ctx->next < BITLIMIT_10 && code < 258);
}

bool
validate_LZW_11bitlitspec(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "11 bit lit: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	uint64_t code = H_CAST_UINT(p->ast);
	return (ctx->next >= BITLIMIT_10 && ctx->next < BITLIMIT_11 && code < 258);
}

bool
validate_LZW_12bitlitspec(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "12 bit lit: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	uint64_t code = H_CAST_UINT(p->ast);
	return (ctx->next >= BITLIMIT_11 && ctx->next < BITLIMIT_12 && code < 258);
}

bool
validate_LZW_clear(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 256);
}

bool
validate_LZW_eod(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 257);
}

bool
validate_LZW_literal(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code < 256);
}

HParsedToken*
act_LZW_literal(const HParseResult *p, void *u)
{
	size_t next_entry_size;
	uint8_t * next_entry_token;
	HBytes * prev_string;
	uint64_t code = H_CAST_UINT(p->ast);
	LZW_context_T * ctx = (LZW_context_T *) u;
	/*
	 * Literals go from 0-255, so they are guaranteed to fit into 1 byte. See also: validate_LZW_literal
	 */
	uint8_t *output = H_ALLOC(uint8_t);
	*output = (uint8_t) code;
	//fprintf(debug, "lit: %lu, next: %u\n", code, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", code); // DEBUG
	//fflush(debug); // DEBUG

	/*
	 * Update the dictionary with the new string. Use of system allocator
	 * here and in act_LZW_codeword is intentional, as LZW_clear_table/init_LZW_context free these
	 */
	prev_string = ctx->lzw_code_table[ctx->old];
	next_entry_size = prev_string->len + 1;
	next_entry_token = calloc(next_entry_size, sizeof(uint8_t));
	memcpy(next_entry_token, prev_string->token, prev_string->len);
	next_entry_token[next_entry_size - 1] = (uint8_t) code;
	lzw_table_insert(ctx, next_entry_token, next_entry_size);

	ctx->old = code;
	return H_MAKE_BYTES(output, 1);
}

HParsedToken*
act_LZW_codeword(const HParseResult *p, void *u)
{
	HBytes * prev_string;
	uint64_t code = H_CAST_UINT(p->ast);
	uint8_t prefix;
	uint8_t * output_token;
	uint8_t * next_entry_token;
	size_t prev_string_length;
	LZW_context_T * ctx = (LZW_context_T *) u;

	//fprintf(debug, "code: %lu, next: %u\n", code, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", code); // DEBUG
	//fflush(debug); // DEBUG


	if(ctx->lzw_code_table[code] != NULL) // code is in the table
	{
		HBytes * code_str;
		size_t code_token_length;

		/*
		 * Retrieve the output from the dictionary.
		 * This is what we'll wrap in a HBytes for returning
		 */
		code_str = ctx->lzw_code_table[code];
		code_token_length = code_str->len;
		output_token = calloc(code_token_length, sizeof(uint8_t));
		memcpy(output_token, code_str->token, code_token_length);

		prev_string = ctx->lzw_code_table[ctx->old];
		prev_string_length = prev_string->len;

		/*
		 * Update the dictionary
		 */
		prefix = output_token[0];
		next_entry_token = calloc(prev_string_length+1, sizeof(uint8_t));
		memcpy(next_entry_token, prev_string->token, prev_string_length);
		next_entry_token[prev_string_length] = prefix;
		lzw_table_insert(ctx, next_entry_token, prev_string_length+1);
		ctx->old = code;

		return H_MAKE_BYTES(output_token, code_token_length);
	}
	else // code is not in the table
	{
		uint8_t new_prefix;
		HBytes * missing_table_entry;
		uint8_t * missing_table_entry_token;
		size_t output_length;

		prev_string = ctx->lzw_code_table[ctx->old];
		prev_string_length = prev_string->len;
		prefix = prev_string->token[0];

		/*
		 * Put together the string for the current code, then insert it into the table. We also copy the token into a separate uint8_t to be returned by the function
		 */
		output_length = prev_string_length + 1;
		output_token = calloc(output_length, sizeof(uint8_t));
		memcpy(output_token, prev_string->token, prev_string_length);
		/*
		 * Output is one byte longer than prev_string, and the last byte is the first character of the previous string
		 */
		output_token[prev_string_length] = prefix;


		missing_table_entry = malloc(sizeof(HBytes));
		missing_table_entry->len = prev_string_length + 1;
		missing_table_entry_token = calloc(missing_table_entry->len, sizeof(uint8_t));
		memcpy(missing_table_entry_token, output_token, missing_table_entry->len);
		missing_table_entry->token = missing_table_entry_token;
		ctx->lzw_code_table[code] = missing_table_entry;

		/*
		 * Update the dictionary
		 */
		new_prefix = output_token[0];
		next_entry_token = calloc(prev_string_length+1, sizeof(uint8_t));
		memcpy(next_entry_token, prev_string->token, prev_string_length);
		next_entry_token[prev_string_length] = new_prefix;
		lzw_table_insert(ctx, next_entry_token, prev_string_length+1);
		ctx->old = code;

		return H_MAKE_BYTES(output_token, output_length);
	}
}

HParsedToken*
act_LZW_body(const HParseResult *p, void *u)
{
	size_t index = 0;
	size_t total_buffer_size = 0;
	size_t num_fragments = h_seq_len(p->ast);
	uint8_t * buffer;

	/* sum total bytes in array, alloc buffer */
	for(int i = 0; i < num_fragments; i++)
	{
		total_buffer_size += H_FIELD_BYTES(i).len;
	}

	buffer = h_arena_malloc(p->arena, sizeof(uint8_t) * total_buffer_size); // XXX arena alloc, calloc

	/* go through parse result, merge bytes */
	for(int i = 0; i < num_fragments; i++)
	{
		size_t len = H_FIELD_BYTES(i).len;
		memcpy(&buffer[index], H_FIELD_BYTES(i).token, len);
		index += len;
	}

	//fprintf(debug, "\n\n"); // DEBUG
	//fwrite(buffer, 1, total_buffer_size, debug); // DEBUG
	//fflush(debug); // DEBUG

	return H_MAKE_BYTES(buffer, total_buffer_size);
}


HParsedToken*
act_LZW_data(const HParseResult *p, void *u)
{
	/* The AST this semantic action receives is a sequence that looks something like this:
		elements[0] -> TT_BYTES representing the initial clear code
		elements[1] -> TT_BYTES representing the first code (should be a literal)
		elements[2] -> TT_BYTES containing the decompressed data (except for the first code)
		elements[3] -> TT_UINT representing the EOD code
		elements[4] -> TT_UINT representing the remaining bits from EOD to the end of the byte, should be 0
	*/

	//HCountedArray * seq = H_CAST_SEQ(p->ast);
	//LZW_context_T *ctx = (LZW_context_T*) u; // DEBUG
	size_t total_buffer_size = 0;
	uint8_t * buffer;
	HBytes first = H_FIELD_BYTES(1);
	HBytes rest = H_FIELD_BYTES(2);

	total_buffer_size = first.len + rest.len;

	buffer = calloc(total_buffer_size, sizeof(uint8_t));
	memcpy(buffer, first.token, first.len);
	memcpy(buffer+first.len, rest.token, rest.len);
	// XXX: Memory use would be greatly decreased if first.token and rest.token could be freed here (allocated in act_LZW_firstcode and act_LZW_body)

	//fprintf(debug, "\n\n"); // DEBUG
	/*for(int i = 258; i < ctx->next; ++i) // DEBUG
	{
		fprintf(debug, "i: %u, str: ", i);
		fwrite(ctx->lzw_code_table[i]->token, ctx->lzw_code_table[i]->len, 1, debug);
		fprintf(debug, "\n");
	}
	fflush(debug); // DEBUG */

	return H_MAKE_BYTES(buffer, total_buffer_size);
}


void init_LZW_parser()
{
	context = malloc(sizeof(LZW_context_T));
	memset(context, 0, sizeof(*context));
	context->next = 258;
	/* set up literals in LZW code table */
	for(int i = 0; i < 256; i++)
	{
		uint8_t *token = malloc(sizeof(uint8_t));
		*token = i;
		HBytes *lit = malloc(sizeof(HBytes));
		lit->token = token;
		lit->len = 1;
		context->lzw_code_table[i] = lit;
	}
	context->earlychange = 1;
	context->old = 257;

	H_VDRULE(LZW_9bitcodeword, h_bits(9, false), context);
	H_VDRULE(LZW_10bitcodeword, h_bits(10, false), context);
	H_VDRULE(LZW_11bitcodeword, h_bits(11, false), context);
	H_VDRULE(LZW_12bitcodeword, h_bits(12, false), context);

	H_VDRULE(LZW_9bitlitspec, h_bits(9, false), context);
	H_VDRULE(LZW_10bitlitspec, h_bits(10, false), context);
	H_VDRULE(LZW_11bitlitspec, h_bits(11, false), context);
	H_VDRULE(LZW_12bitlitspec, h_bits(12, false), context);

	H_RULE(LZW_remainingbits, h_many(h_bits(1, false))); //XXX: could validate that these bits are 0?

	H_ADRULE(LZW_firstcode, LZW_9bitlitspec, context); // First code is always a literal, sets ctx->old

	H_AVDRULE(LZW_clear, h_choice(LZW_9bitlitspec, LZW_10bitlitspec, LZW_11bitlitspec, LZW_12bitlitspec, NULL), context);
	H_VDRULE(LZW_eod, h_choice(LZW_9bitlitspec, LZW_10bitlitspec, LZW_11bitlitspec, LZW_12bitlitspec, NULL), context);
	H_AVDRULE(LZW_literal, h_choice(LZW_9bitlitspec, LZW_10bitlitspec, LZW_11bitlitspec, LZW_12bitlitspec, NULL), context);
	H_ADRULE(LZW_codeword, h_choice(LZW_9bitcodeword, LZW_10bitcodeword, LZW_11bitcodeword, LZW_12bitcodeword, NULL), context);

	H_ADRULE(LZW_clear_and_firstcode, h_sequence(h_ignore(LZW_clear), LZW_firstcode, NULL), context);

	H_ADRULE(LZW_body, h_many1(h_butnot(h_choice(LZW_literal, LZW_clear_and_firstcode, LZW_codeword, NULL), LZW_eod)), context);

	H_ADRULE(LZW_data, h_sequence(LZW_clear, LZW_firstcode, LZW_body, LZW_eod, LZW_remainingbits, NULL), context);
	p_lzwdata = LZW_data;
}


HParseResult* parse_LZW_data(const uint8_t* input, size_t length)
{
	//debug = fopen("lzw_debug.txt", "a"); // DEBUG
	HParseResult *res = h_parse(p_lzwdata, input, length);
	//fclose(debug); // DEBUG
	return res;
}


void init_LZW_context(int earlychange)
{
	for(int i = 258; i < 4096; ++i)
	{
		if(context->lzw_code_table[i] != NULL)
		{
			free((uint8_t *) context->lzw_code_table[i]->token); // These can be freed without issue, because HParsedTokens containing them have separate deep copies
			free(context->lzw_code_table[i]);
		}
		context->lzw_code_table[i] = NULL;
	}
	context->next = 258;
	context->old = 257;
	context->earlychange = earlychange;
}
